import warnings

import numpy as np

from rubin_scheduler.scheduler.utils import smallest_signed_angle
from rubin_scheduler.utils import (
    Site,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    approx_alt_az2_ra_dec,
    rotation_converter,
)

from .jerk import jerk_time

__all__ = ("KinemModel",)
two_pi = 2.0 * np.pi


class Radec2altazpa:
    """Class to make it easy to swap in different alt/az conversion if
    wanted."""

    def __init__(self, location):
        self.location = location

    def __call__(self, ra, dec, mjd):
        alt, az, pa = _approx_ra_dec2_alt_az(
            ra, dec, self.location.lat_rad, self.location.lon_rad, mjd, return_pa=True
        )
        return alt, az, pa


class KinemModel:
    """A Kinematic model of the telescope.

    Parameters
    ----------
    location : `astropy.coordinates.EarthLocation`
        The location of the telescope.
        If None, defaults to rubin_scheduler.utils.Site info
    park_alt : `float` (86.5)
        The altitude the telescope gets parked at (degrees)
    park_az : `float` (0)
        The azimuth for telescope park position (degrees)
    start_filter : `str` ('r')
        The filter that gets loaded when the telescope is parked
    mjd0 : `float` (0)
        The MJD to assume we are starting from
    telescope : `str'
        The telescope name to use for sky rotation conversion. Default "rubin"

    Note
    ----
    Note there are additional parameters in the methods setup_camera,
    setup_dome, setup_telescope, and setup_optics.
    Just breaking it up a bit to make it more readable.
    """

    def __init__(
        self, location=None, park_alt=86.5, park_az=0.0, start_filter="r", mjd0=0, telescope="rubin"
    ):
        self.park_alt_rad = np.radians(park_alt)
        self.park_az_rad = np.radians(park_az)
        self.start_filter = start_filter
        self.current_filter = self.start_filter
        if location is None:
            self.location = Site("LSST")
            self.location.lat_rad = np.radians(self.location.latitude)
            self.location.lon_rad = np.radians(self.location.longitude)
        # Our RA,Dec to Alt,Az converter
        self.radec2altaz = Radec2altazpa(self.location)

        self.setup_camera()
        self.setup_dome()
        self.setup_telescope()
        self.setup_optics()

        # Park the telescope
        self.park()
        self.last_mjd = mjd0

        # Rotation conversion
        self.rc = rotation_converter(telescope=telescope)

    def mount_filters(self, filter_list):
        """Change which filters are mounted

        Parameters
        ----------
        filter_list : `list` [`str`]
            List of the mounted filters.
        """
        self.mounted_filters = filter_list

    def setup_camera(
        self,
        readtime=2.0,
        shuttertime=1.0,
        filter_changetime=120.0,
        fov=3.5,
        rotator_min=-90,
        rotator_max=90,
        maxspeed=3.5,
        accel=1.0,
        shutter_2motion_min_time=15.0,
        jerk=None,
    ):
        """
        Parameters
        ----------
        readtime : `float` (2)
            The readout time of the CCDs (seconds)
        shuttertime : `float` (1.)
            The time it takes the shutter to go from closed to fully open
            (seconds)
        filter_changetime : `float` (120)
            The time it takes to change filters (seconds)
        fov : `float` (3.5)
            The camera field of view (degrees)
        rotator_min : `float` (-90)
            The minimum angle the camera rotator (rotTelPos) can move to
            (degrees)
        rotator_max : `float` (90)
            The maximum angle the camera rotator (rotTelPos) can move to
            degrees)
        maxspeed : `float` (3.5)
            The maximum speed of the rotator (degrees/s)
        accel : `float` (1.0)
            The acceleration of the rotator (degrees/s^2)
        jerk : `float`
            The jerk of the rotator (degrees/s^3). Default of None treats
            jerk as infinite
        shutter_2motion_min_time : `float` (15.)
            The time required for two shutter motions (seconds). If one
            takes a 1-snap 10s exposure, there will be a 5s of overhead
            before the next exposure can start.
        """
        self.readtime = readtime
        self.shuttertime = shuttertime
        self.filter_changetime = filter_changetime
        self.camera_fov = np.radians(fov)

        self.telrot_minpos_rad = np.radians(rotator_min)
        self.telrot_maxpos_rad = np.radians(rotator_max)
        self.telrot_maxspeed_rad = np.radians(maxspeed)
        self.telrot_accel_rad = np.radians(accel)
        self.telrot_jerk_rad = np.radians(jerk) if jerk is not None else None
        self.shutter_2motion_min_time = shutter_2motion_min_time
        self.mounted_filters = ["u", "g", "r", "i", "y"]

    def setup_dome(
        self,
        altitude_maxspeed=1.75,
        altitude_accel=0.875,
        altitude_jerk=None,
        altitude_freerange=0.0,
        azimuth_maxspeed=1.5,
        azimuth_accel=0.75,
        azimuth_jerk=None,
        azimuth_freerange=4.0,
        settle_time=1.0,
    ):
        """Parameters to define the DOME movement.

        Parameters
        ----------
        altitude_maxspeed : `float` (1.75)
            Maximum speed for altitude movement (degrees/second)
        altitude_accel : `float` (0.875)
            Maximum acceleration for altitude movement (degrees/second**2)
        altitude_jerk : `float`
            The jerk for the altitude movement (degrees/second**3). Default
            of None treats jerk as infinite.
        altitude_freerange : `float` (0)
            The range over which there is 0 delay
        azimuth_maxspeed : `float` (1.5)
            Maximum speed for azimuth movement (degrees/second)
        azimuth_accel : `float` (0.75)
            Maximum acceleration for azimuth movement (degrees/second**2)
        azimuth_jerk : `float`
            The jerk of the azimuth movement (degrees/second**3). Default
            of None treats jerk as infinite.
        azimuth_freerange : `float` (4.0)
            The range in which there is 0 delay
        settle_time : `float` (1.0)
            Settle time after movement (seconds)
        """
        self.domalt_maxspeed_rad = np.radians(altitude_maxspeed)
        self.domalt_accel_rad = np.radians(altitude_accel)
        self.domalt_jerk_rad = np.radians(altitude_jerk) if altitude_jerk is not None else None
        self.domalt_free_range = np.radians(altitude_freerange)
        self.domaz_maxspeed_rad = np.radians(azimuth_maxspeed)
        self.domaz_accel_rad = np.radians(azimuth_accel)
        self.domaz_jerk_rad = np.radians(azimuth_jerk) if azimuth_jerk is not None else None
        self.domaz_free_range = np.radians(azimuth_freerange)
        self.domaz_settletime = settle_time

    def setup_telescope(
        self,
        altitude_minpos=20.0,
        altitude_maxpos=86.5,
        azimuth_minpos=-250.0,
        azimuth_maxpos=250.0,
        altitude_maxspeed=3.5,
        altitude_accel=3.5,
        altitude_jerk=None,
        azimuth_maxspeed=7.0,
        azimuth_accel=7.0,
        azimuth_jerk=None,
        settle_time=3.0,
        az_limits=None,
        alt_limits=None,
    ):
        """Parameters to define the TELESCOPE movement and position.

        Parameters
        ----------
        altitude_minpos : `float` (20.0)
            Minimum altitude for the telescope (degrees)
        altitude_maxpos : `float` (86.5)
            Maximum altitude for the telescope (degrees)
        azimuth_minpos : `float` (-270.0)
            Minimum azimuth position (degrees)
        azimuth_maxpos : `float` (270.0)
            Maximum azimuth position (degrees)
        altitude_maxspeed : `float` (3.5)
            Maximum speed for altitude movement (degrees/second)
        altitude_accel : `float` (3.5)
            Maximum acceleration for altitude movement (degrees/second**2)
        azimuth_maxspeed : `float` (7.0)
            Maximum speed for azimuth movement (degrees/second)
        azimuth_accel : `float` (7.0)
            Maximum acceleration for azimuth movement (degrees/second**2)
        settle_time : `float` (3.0)
            Settle time required for telescope after movement (seconds)
        """
        self.telalt_minpos_rad = np.radians(altitude_minpos)
        self.telalt_maxpos_rad = np.radians(altitude_maxpos)
        self.telaz_minpos_rad = np.radians(azimuth_minpos)
        self.telaz_maxpos_rad = np.radians(azimuth_maxpos)
        self.telalt_maxspeed_rad = np.radians(altitude_maxspeed)
        self.telalt_accel_rad = np.radians(altitude_accel)
        self.telalt_jerk_rad = np.radians(altitude_jerk) if altitude_jerk is not None else None
        self.telaz_maxspeed_rad = np.radians(azimuth_maxspeed)
        self.telaz_accel_rad = np.radians(azimuth_accel)
        self.telaz_jerk_rad = np.radians(azimuth_jerk) if azimuth_jerk is not None else None
        self.mount_settletime = settle_time
        if alt_limits is None:
            self.alt_limits = [[self.telalt_minpos_rad, self.telalt_maxpos_rad]]
        else:
            self.alt_limits = np.radians(alt_limits)
        if az_limits is None:
            self.az_limits = [[0, 2.0 * np.pi]]
        else:
            self.az_limits = np.radians(az_limits)

    def setup_optics(self, ol_slope=1.0 / 3.5, cl_delay=[0.0, 36.0], cl_altlimit=[0.0, 9.0, 90.0]):
        """
        Parameters
        ----------
        ol_slope : `float` (1.0/3.5)
            seconds/degree in altitude slew.
        cl_delay : list ([0.0, 36])
            The delays for closed optics loops (seconds)
        cl_altlimit : list ([0.0, 9.0, 90.0])
            The altitude limits (degrees) for performing closed optice loops.
            Should be one element longer than cl_delay.

        Note
        ----
        A given movement in altitude will cover X degrees;
        if X > cl_altlimit[i] there is
        an additional delay of cl_delay[i]
        """
        # ah, 1./np.radians(1)=np.pi/180
        self.optics_ol_slope = ol_slope / np.radians(1.0)
        self.optics_cl_delay = cl_delay
        self.optics_cl_altlimit = np.radians(cl_altlimit)

    def park(self):
        """Put the telescope in the park position."""
        # I'm going to ignore that the old model had the dome altitude at 90
        # and telescope altitude 86 for park.
        # We should usually be dome az limited anyway, so this should be a
        # negligible approximation.
        self.parked = True

        # We have no current position we are tracking
        self.current_ra_rad = None
        self.current_dec_rad = None
        self.current_rot_sky_pos_rad = None
        self.cumulative_azimuth_rad = 0

        # The last position we were at (or the current if we are parked)
        self.last_az_rad = self.park_az_rad
        self.last_alt_rad = self.park_alt_rad
        self.last_rot_tel_pos_rad = 0

        # Any overhead that must happen before next exposure can start. Slew
        # motions are allowed during the overhead time
        self.overhead = 0.0

        # Don't leave random filter in overnight
        self.current_filter = self.start_filter

    def current_alt_az(self, mjd):
        """return the current alt az position that we have tracked to."""
        if self.parked:
            return self.last_alt_rad, self.last_az_rad, self.last_rot_tel_pos_rad
        else:
            alt_rad, az_rad, pa = self.radec2altaz(self.current_ra_rad, self.current_dec_rad, mjd)
            rot_tel_pos = self.rc._rotskypos2rottelpos(self.current_rot_sky_pos_rad, pa)
            return alt_rad, az_rad, rot_tel_pos

    def slew_times(
        self,
        ra_rad,
        dec_rad,
        mjd,
        rot_sky_pos=None,
        rot_tel_pos=None,
        filtername="r",
        lax_dome=True,
        alt_rad=None,
        az_rad=None,
        starting_alt_rad=None,
        starting_az_rad=None,
        starting_rot_tel_pos_rad=None,
        update_tracking=False,
    ):
        """Calculates slew time to a series of alt/az/filter positions
        from the current position (stored internally).

        Assumptions (currently):
        Assumes we have been tracking on ra,dec,rot_sky_pos position.
        Ignores the motion of the sky while we are slewing
        (this approx should probably average out over time).
        No checks for if we have tracked beyond limits.
        (this assumes folks put telescope in park if there's a long gap.)
        Assumes the camera rotator never needs to (or can't) do a slew
        over 180 degrees.

        Calculates the slew time necessary to get from current state
        to alt2/az2/filter2. The time returned is actually the time between
        the end of an exposure at current location and the beginning of an
        exposure at alt2/az2, since it includes readout time in
        the slew time.

        Parameters
        ----------
        ra_rad : `np.ndarray`
            The RA(s) of the location(s) we wish to slew to (radians)
        dec_rad : `np.ndarray`
            The declination(s) of the location(s) we wish to slew to (radians)
        mjd : `float`
            The current moodified julian date (days)
        rot_sky_pos : `np.ndarray`
            The desired rot_sky_pos(s) (radians).
            Angle between up on the chip and North.
            Note, it is possible to set a rot_sky_pos outside the allowed
            camera rotator range, in which case the slewtime will be np.inf.
            If both rot_sky_pos and rot_tel_pos are set,
            rot_tel_pos will be used.
        rot_tel_pos : `np.ndarray`
            The desired rot_tel_pos(s) (radians).
        filtername : `str`
            The filter(s) of the desired observations.
            Set to None to compute only telescope and dome motion times.
        alt_rad : `np.ndarray`
            The altitude(s) of the destination pointing(s) (radians).
            Will override ra_rad,dec_rad if provided.
        az_rad : `np.ndarray`
            The azimuth(s) of the destination pointing(s) (radians).
            Will override ra_rad,dec_rad if provided.
        lax_dome : `bool`, default True
            If True, allow the dome to creep, model a dome slit, and don't
            require the dome to settle in azimuth. If False, adhere to the
            way SOCS calculates slew times (as of June 21 2017) and do not
            allow dome creep.
        starting_alt_rad : `float` (None)
            The starting altitude for the slew (radians).
            If None, will use internally stored last pointing.
        starting_az_rad : `float` (None)
            The starting azimuth for the slew (radians).
            If None, will use internally stored last pointing.
        starting_rot_tel_pos_rad : `float` (None)
            The starting camera rotation for the slew (radians).
            If None, will use internally stored last pointing.
        update_tracking : `bool` (False)
            If True, update the internal attributes to say we are tracking
            the specified RA,Dec,RotSkyPos position.

        Returns
        -------
        slew_time : `np.ndarray`
            The number of seconds between the two specified exposures.
            Will be np.nan or np.inf if slew is not possible.
        """
        if filtername not in self.mounted_filters:
            return np.nan

        # Don't trust folks to do pa calculation correctly, if both
        # rotations set, rot_sky_pos wins
        if (rot_tel_pos is not None) & (rot_sky_pos is not None):
            if np.isfinite(rot_tel_pos):
                rot_sky_pos = None
            else:
                rot_tel_pos = None

        # alt,az not provided, calculate from RA,Dec
        if alt_rad is None:
            alt_rad, az_rad, pa = self.radec2altaz(ra_rad, dec_rad, mjd)
        else:
            pa = _approx_altaz2pa(alt_rad, az_rad, self.location.lat_rad)
            if update_tracking:
                ra_rad, dec_rad = approx_alt_az2_ra_dec(
                    alt_rad, az_rad, self.location.lat_rad, self.location.lon_rad, mjd
                )

        if starting_alt_rad is None:
            if self.parked:
                starting_alt_rad = self.park_alt_rad
                starting_az_rad = self.park_az_rad
            else:
                starting_alt_rad, starting_az_rad, starting_pa = self.radec2altaz(
                    self.current_ra_rad, self.current_dec_rad, mjd
                )

        delta_alt = np.abs(alt_rad - starting_alt_rad)
        delta_az_short = smallest_signed_angle(starting_az_rad, az_rad)
        delta_az_long = delta_az_short - two_pi
        daslz = np.where(delta_az_short < 0)[0]
        delta_az_long[daslz] = two_pi + delta_az_short[daslz]
        # So, for every position, we can get there by slewing long or
        # short way
        cummulative_az_short = delta_az_short + self.cumulative_azimuth_rad
        oob = np.where(
            (cummulative_az_short < self.telaz_minpos_rad) | (cummulative_az_short > self.telaz_maxpos_rad)
        )[0]
        # Set out of bounds azimuths to infinite distance
        delta_az_short[oob] = np.inf
        cummulative_az_long = delta_az_long + self.cumulative_azimuth_rad
        oob = np.where(
            (cummulative_az_long < self.telaz_minpos_rad) | (cummulative_az_long > self.telaz_maxpos_rad)
        )[0]
        delta_az_long[oob] = np.inf

        # Find minimum azimuth slew out of long/short direction (use
        # absolute, because these can be negative)
        # Note that with an impaired telescope with az range<180,
        # infinite slewtimes can propagate
        delta_aztel = np.where(
            np.abs(delta_az_short) < np.abs(delta_az_long),
            delta_az_short,
            delta_az_long,
        )

        # Calculate how long the telescope will take to slew to this
        # position.
        tel_alt_slew_time = jerk_time(
            delta_alt, self.telalt_maxspeed_rad, self.telalt_accel_rad, self.telalt_jerk_rad
        )
        tel_az_slew_time = jerk_time(
            np.abs(delta_aztel), self.telaz_maxspeed_rad, self.telaz_accel_rad, self.telaz_jerk_rad
        )
        tot_tel_time = np.maximum(tel_alt_slew_time, tel_az_slew_time)

        # Time for open loop optics correction
        ol_time = delta_alt / self.optics_ol_slope
        tot_tel_time += ol_time
        # Add time for telescope settle.
        # note, this means we're going to have a settle time even for very
        # small slews like dithering.
        settle_and_ol = np.where(tot_tel_time > 0)
        tot_tel_time[settle_and_ol] += np.maximum(0, self.mount_settletime - ol_time[settle_and_ol])

        # And any leftover overhead sets a minimum on the total telescope
        # time
        tot_tel_time = np.maximum(self.overhead, tot_tel_time)

        # now compute dome slew time
        # the dome can spin all the way around, so we will let it go the
        # shortest angle, even if the telescope has to unwind
        delta_az = np.abs(smallest_signed_angle(starting_az_rad, az_rad))
        if lax_dome:
            # model dome creep, dome slit, and no azimuth settle
            # if we can fit both exposures in the dome slit, do so
            same_dome = np.where(delta_alt**2 + delta_az**2 < self.camera_fov**2)

            # else, we take the minimum time from two options:
            # 1. assume we line up alt in the center of the dome slit so we
            #    minimize distance we have to travel in azimuth.
            # 2. line up az in the center of the slit
            # also assume:
            # * that we start out going maxspeed for both alt and az
            # * that we only just barely have to get the new field in the
            #   dome slit in one direction, but that we have to center the
            #   field in the other (which depends which of the two options
            #   used)
            # * that we don't have to slow down until after the shutter
            #   starts opening
            dom_delta_alt = delta_alt
            # on each side, we can start out with the dome shifted away from
            # the center of the field by an amount domSlitRadius - fovRadius
            dom_slit_diam = self.camera_fov / 2.0
            dom_delta_az = delta_az - 2 * (dom_slit_diam / 2 - self.camera_fov / 2)
            dom_alt_slew_time = dom_delta_alt / self.domalt_maxspeed_rad
            dom_az_slew_time = dom_delta_az / self.domaz_maxspeed_rad
            tot_dom_time1 = np.maximum(dom_alt_slew_time, dom_az_slew_time)

            dom_delta_alt = delta_alt - 2 * (dom_slit_diam / 2 - self.camera_fov / 2)
            dom_delta_az = delta_az
            dom_alt_slew_time = dom_delta_alt / self.domalt_maxspeed_rad
            dom_az_slew_time = dom_delta_az / self.domaz_maxspeed_rad
            tot_dom_time2 = np.maximum(dom_alt_slew_time, dom_az_slew_time)

            tot_dom_time = np.minimum(tot_dom_time1, tot_dom_time2)
            tot_dom_time[same_dome] = 0

        else:
            # the above models a dome slit and dome creep. However, it
            # appears that SOCS requires the dome to slew exactly to each
            # field and settle in az
            dom_alt_slew_time = jerk_time(
                delta_alt, self.domalt_maxspeed_rad, self.domalt_accel_rad, self.domalt_jerk_rad
            )
            dom_az_slew_time = jerk_time(
                delta_az, self.domaz_maxspeed_rad, self.domaz_accel_rad, self.domaz_jerk_rad
            )

            # Dome takes 1 second to settle in az
            dom_az_slew_time = np.where(
                dom_az_slew_time > 0,
                dom_az_slew_time + self.domaz_settletime,
                dom_az_slew_time,
            )
            tot_dom_time = np.maximum(dom_alt_slew_time, dom_az_slew_time)
        # Find the max of the above for slew time.
        slew_time = np.maximum(tot_tel_time, tot_dom_time)
        # include filter change time if necessary. Assume no filter
        # change time needed if we are starting parked
        if not self.parked:
            filter_change = np.where(filtername != self.current_filter)
            slew_time[filter_change] = np.maximum(slew_time[filter_change], self.filter_changetime)
        # Add closed loop optics correction
        # Find the limit where we must add the delay
        cl_limit = self.optics_cl_altlimit[1]
        cl_delay = self.optics_cl_delay[1]
        close_loop = np.where(delta_alt >= cl_limit)
        slew_time[close_loop] += cl_delay

        # Mask min/max altitude limits so slewtime = np.nan
        outside_limits = np.where((alt_rad > self.telalt_maxpos_rad) | (alt_rad < self.telalt_minpos_rad))[0]
        slew_time[outside_limits] = np.nan

        # If we want to include the camera rotation time
        if (rot_sky_pos is not None) | (rot_tel_pos is not None):
            if rot_tel_pos is None:
                # This is now between -pi and pi
                rot_tel_pos = self.rc._rotskypos2rottelpos(rot_sky_pos, pa)
            if rot_sky_pos is None:
                rot_sky_pos = self.rc._rottelpos2rotskypos(rot_tel_pos, pa)

            # Is the new rot_tel_pos reachable? If not return NaN
            if (rot_tel_pos < self.telrot_minpos_rad) | (rot_tel_pos > self.telrot_maxpos_rad):
                return np.nan
            # If there was no kwarg for starting rotator position
            if starting_rot_tel_pos_rad is None:
                # If there is no current rot_sky_pos, we were parked
                if self.current_rot_sky_pos_rad is None:
                    current_rot_tel_pos = self.last_rot_tel_pos_rad
                else:
                    # We have been tracking, so rot_tel_pos needs to be updated
                    current_rot_tel_pos = self.rc._rotskypos2rottelpos(
                        self.current_rot_sky_pos_rad, starting_pa
                    )
            else:
                # kwarg overrides if it was supplied
                current_rot_tel_pos = starting_rot_tel_pos_rad
            delta_rotation = np.abs(smallest_signed_angle(current_rot_tel_pos, rot_tel_pos))
            rotator_time = jerk_time(
                delta_rotation, self.telrot_maxspeed_rad, self.telrot_accel_rad, self.telrot_jerk_rad
            )
            slew_time = np.maximum(slew_time, rotator_time)

        # Update the internal attributes to note that we are now pointing
        # and tracking at the requested RA,Dec,rot_sky_pos
        if update_tracking:
            self.current_ra_rad = ra_rad
            self.current_dec_rad = dec_rad
            self.current_rot_sky_pos_rad = rot_sky_pos
            self.parked = False
            # Handy to keep as reference, but not used for any calculations
            self.last_rot_tel_pos_rad = rot_tel_pos
            self.last_az_rad = az_rad
            self.last_alt_rad = alt_rad
            self.last_pa_rad = pa
            # Track the cumulative azimuth
            self.cumulative_azimuth_rad += delta_aztel
            self.current_filter = filtername
            self.last_mjd = mjd

        return slew_time

    def visit_time(self, observation):
        # How long does it take to make an observation.
        visit_time = (
            observation["exptime"]
            + observation["nexp"] * self.shuttertime
            + max(observation["nexp"] - 1, 0) * self.readtime
        )
        return visit_time

    def shutter_stall(self, observation):
        """Time we need to stall after shutter closes to let things cool
        down."""
        result = 0.0
        delta_t = observation["exptime"] / observation["nexp"]
        if delta_t < self.shutter_2motion_min_time:
            result = self.shutter_2motion_min_time - delta_t
        return result

    def observe(self, observation, mjd, rot_tel_pos=None, lax_dome=True):
        """Observe a target, and return the slewtime and visit time for
        the action

        If slew is not allowed, returns np.nan and does not update state.
        """
        if (observation["nexp"] >= 2) & (
            observation["exptime"] / observation["nexp"] < self.shutter_2motion_min_time
        ):
            msg = "%i exposures in %i seconds is violating number of shutter motion limit" % (
                observation["nexp"],
                observation["exptime"],
            )
            warnings.warn(msg)

        slewtime = self.slew_times(
            observation["RA"],
            observation["dec"],
            mjd,
            rot_sky_pos=observation["rotSkyPos"],
            rot_tel_pos=rot_tel_pos,
            filtername=observation["filter"],
            update_tracking=True,
            lax_dome=lax_dome,
        )
        visit_time = self.visit_time(observation)
        # Compute any overhead that is left over from this
        if ~np.isnan(slewtime):
            self.overhead = np.maximum(self.readtime, self.shutter_stall(observation))

        return slewtime, visit_time



__all__ = (
    "TargetMapModuloBasisFunction",
    "FootprintBasisFunction",
    "FootprintRollingBasisFunction",
)

import warnings

import healpy as hp
import numpy as np

from rubin_scheduler.scheduler import features, utils
from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import ConstantFootprint


def send_unused_deprecation_warning(name):
    message = (
        f"The basis function {name} is not in use by the current "
        "baseline scheduler and may be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class FootprintBasisFunction(BaseBasisFunction):
    """Basis function that tries to maintain a uniformly covered footprint

    Parameters
    ----------
    filtername : `str`, optional
        The filter for this footprint. Default r.
    nside : `int`, optional
        The nside for the basis function and features.
        Default None uses `~utils.set_default_nside()`
    footprint : `~rubin_scheduler.scheduler.utils.Footprint` object
        The desired footprint. The default will set this to None,
        but in general this is really not desirable.
        In order to make default a kwarg, a current baseline footprint
        is setup with a Constant footprint (not rolling, not even season
        aware).
    out_of_bounds_val : `float`, optional
        The value to set the basis function for regions that are not in
        the footprint. Default -10, np.nan is another good value to use.

    """

    def __init__(
        self,
        filtername="r",
        nside=None,
        footprint=None,
        out_of_bounds_val=-10.0,
    ):
        super().__init__(nside=nside, filtername=filtername)
        if footprint is None:
            # This is useful as a backup, but really footprint SHOULD
            # be specified when basis function is set up.
            # This just uses whole survey, but doesn't set up rolling.
            warnings.warn("No Footprint set, using a constant default.")
            target_maps, labels = utils.get_current_footprint(self.nside)
            fp = ConstantFootprint(nside=self.nside)
            for f in "ugrizy":
                fp.set_footprint(f, target_maps[f])
        self.footprint = footprint

        self.survey_features = {}
        # All the observations in all filters
        self.survey_features["N_obs_all"] = features.NObservations(nside=self.nside, filtername=None)
        self.survey_features["N_obs"] = features.NObservations(nside=self.nside, filtername=filtername)

        # should probably actually loop over all the target maps?
        self.out_of_bounds_area = np.where(footprint.get_footprint(self.filtername) <= 0)[0]
        self.out_of_bounds_val = out_of_bounds_val

    def _calc_value(self, conditions, indx=None):
        # Find out what the footprint object thinks we should have been
        # observed
        desired_footprint_normed = self.footprint(conditions.mjd)[self.filtername]

        # Compute how many observations we should have on the sky
        desired = desired_footprint_normed * np.sum(self.survey_features["N_obs_all"].feature)
        result = desired - self.survey_features["N_obs"].feature
        result[self.out_of_bounds_area] = self.out_of_bounds_val
        return result


class FootprintRollingBasisFunction(BaseBasisFunction):
    """Let's get the rolling really right.

    Parameters
    ----------
    footprints : list of np.array
        List of HEALpix arrays. The footprints should all have matching
        sums and have the same nside.
    all_footprints_sum : float
        The sum of footprints over all filters.
    all_rolling_sum : float
        The sum (over all filters) of the region of the maps that changs.
    season_modulo : int (2)
        The modulo to pass to utils.season_calc.
    season_length : float (365.25)
        How long a season should be (days).
    max_season : int (None)
        If set, the season calc will return -1 for values greater than
        max_season
    day_offset : np.array (None)
        Offset to pass to utils.season_calc (days).

    """

    def __init__(
        self,
        filtername="r",
        nside=None,
        footprints=None,
        all_footprints_sum=None,
        all_rolling_sum=None,
        out_of_bounds_val=-10,
        season_modulo=2,
        season_length=365.25,
        max_season=None,
        day_offset=None,
    ):
        super(FootprintRollingBasisFunction, self).__init__(nside=nside, filtername=filtername)

        # OK, going to find the parts of the map that are the same
        # everywhere, and compute the basis function the same as usual
        # for those.
        same_footprint = np.ones(footprints[0].size, dtype=bool)
        for footprint in footprints[0:-1]:
            same_footprint = same_footprint & (footprint == footprints[-1])

        sum_footprints = footprints[0] * 0
        for footprint in footprints:
            sum_footprints += footprint
        self.constant_footprint_indx = np.where((same_footprint == True) & (sum_footprints > 0))[0]
        self.rolling_footprint_indx = np.where((same_footprint == False) & (sum_footprints > 0))[0]

        self.season_modulo = season_modulo
        self.season_length = season_length
        self.max_season = max_season
        self.day_offset = day_offset
        self.footprints = footprints

        self.max_day_offset = np.max(self.day_offset)

        self.all_footprints_sum = all_footprints_sum
        self.all_rolling_sum = all_rolling_sum

        self.survey_features = {}
        # Set a season for -1 (for before rolling or after max_season)
        # All the observations in the given filter
        self.survey_features["N_obs_%i" % -1] = features.N_observations(nside=nside, filtername=filtername)
        # All the observations in all filters
        self.survey_features["N_obs_all_%i" % -1] = features.N_observations(nside=nside, filtername=None)

        for i, temp in enumerate(footprints[0:-1]):
            # Observation in a season, in filtername
            self.survey_features["N_obs_%i" % i] = features.N_observations_season(
                i,
                filtername=filtername,
                nside=self.nside,
                modulo=season_modulo,
                offset=day_offset,
                max_season=max_season,
                season_length=season_length,
            )
            # Count of all the observations taken in a season
            self.survey_features["N_obs_all_%i" % i] = features.N_observations_season(
                i,
                filtername=None,
                modulo=season_modulo,
                offset=day_offset,
                nside=self.nside,
                max_season=max_season,
                season_length=season_length,
            )

        # Now I need to track the observations taken in each season.
        self.out_of_bounds_area = np.where(footprint <= 0)[0]
        self.out_of_bounds_val = out_of_bounds_val

        self.result = np.zeros(hp.nside2npix(nside), dtype=float)
        send_unused_deprecation_warning(self.__class__.__name__)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()

        # Compute what season it is at each pixel
        seasons = conditions.season.copy()
        seasons = seasons * 365.25 / self.season_length
        if self.max_season is not None:
            over_indx = np.where(seasons >= self.max_season)
        if self.season_modulo is not None:
            neg = np.where(seasons < 0)
            seasons = seasons % self.season_modulo
            seasons[neg] = -1
        if self.max_season is not None:
            seasons[over_indx] = -1

        # Update the coverage of the sky so far
        # If RA to sun is zero, we are at phase np.pi/2.
        coverage_map_phase = (conditions.ra - conditions.sun_RA_start + np.pi / 2.0) % (2.0 * np.pi)
        zero = utils.step_line(0.0, 1.0, 365.25, phase=coverage_map_phase * 365.25 / 2 / np.pi)
        t_elapsed = conditions.mjd - conditions.mjd_start
        norm_coverage_raw = utils.step_line(
            t_elapsed, 1.0, 365.25, phase=coverage_map_phase * 365.25 / 2 / np.pi
        )
        norm_coverage_raw -= zero

        norm_coverage = norm_coverage_raw / np.max(norm_coverage_raw)
        norm_footprint = self.footprints[-1] * norm_coverage

        # Compute the constant parts of the footprint like before
        desired = (
            norm_footprint / self.all_footprints_sum * np.sum(self.survey_features["N_obs_all_-1"].feature)
        )
        result[self.constant_footprint_indx] = (
            desired[self.constant_footprint_indx]
            - self.survey_features["N_obs_-1"].feature[self.constant_footprint_indx]
        )

        # Now for the rolling sections
        norm_coverage = 0
        ns = np.floor((t_elapsed + self.max_day_offset) / 365.25)
        if np.max(ns) > -1:
            ns_on = ns / self.season_modulo
            ns_off = ns - ns_on
            norm_coverage = norm_coverage_raw - ns_off
            norm_coverage = norm_coverage / np.max(norm_coverage)

        for season in np.unique(seasons[self.rolling_footprint_indx]):
            if season == -1:
                nf = norm_footprint
            else:
                nf = self.footprints[season] * norm_coverage
            season_indx = np.where(seasons[self.rolling_footprint_indx] == season)[0]
            desired = (
                nf[self.rolling_footprint_indx[season_indx]]
                / self.all_rolling_sum
                * np.sum(self.survey_features["N_obs_all_%i" % season].feature[self.rolling_footprint_indx])
            )
            result[self.rolling_footprint_indx[season_indx]] = (
                desired
                - self.survey_features["N_obs_%i" % season].feature[self.rolling_footprint_indx][season_indx]
            )

        result[self.out_of_bounds_area] = self.out_of_bounds_val
        return result


class TargetMapModuloBasisFunction(BaseBasisFunction):
    """Basis function that tracks number of observations and tries to match
    a specified spatial distribution can enter multiple maps that will be
    used at different times in the survey

    Parameters
    ----------
    day_offset : np.array
        Healpix map that has the offset to be applied to each pixel when
        computing what season it is on.
    filtername : (string 'r')
        The name of the filter for this target map.
    nside: int (default_nside)
        The healpix resolution.
    target_maps : list of numpy array (None)
        healpix maps showing the ratio of observations desired for all
        points on the sky. Last map will be used for season -1. Probably
        shouldn't support going to season less than -1.
    norm_factor : float (0.00010519)
        for converting target map to number of observations. Should be the
        area of the camera divided by the area of a healpixel divided by
        the sum of all your goal maps. Default value assumes LSST foV has
        1.75 degree radius and the standard goal maps. If using mulitple
        filters, see rubin_scheduler.scheduler.utils.calc_norm_factor for
        a utility that computes norm_factor.
    out_of_bounds_val : float (-10.)
        Reward value to give regions where there are no observations
        requested (unitless).
    season_modulo : int (2)
        The value to modulate the season by (years).
    max_season : int (None)
        For seasons higher than this value (pre-modulo), the final target
        map is used.

    """

    def __init__(
        self,
        day_offset=None,
        filtername="r",
        nside=None,
        target_maps=None,
        norm_factor=None,
        out_of_bounds_val=-10.0,
        season_modulo=2,
        max_season=None,
        season_length=365.25,
    ):
        super(TargetMapModuloBasisFunction, self).__init__(nside=nside, filtername=filtername)

        if norm_factor is None:
            warnings.warn("No norm_factor set, use utils.calc_norm_factor if using multiple filters.")
            self.norm_factor = 0.00010519
        else:
            self.norm_factor = norm_factor

        self.survey_features = {}
        # Map of the number of observations in filter

        for i, temp in enumerate(target_maps[0:-1]):
            self.survey_features["N_obs_%i" % i] = features.N_observations_season(
                i,
                filtername=filtername,
                nside=self.nside,
                modulo=season_modulo,
                offset=day_offset,
                max_season=max_season,
                season_length=season_length,
            )
            # Count of all the observations taken in a season
            self.survey_features["N_obs_count_all_%i" % i] = features.N_obs_count_season(
                i,
                filtername=None,
                season_modulo=season_modulo,
                offset=day_offset,
                nside=self.nside,
                max_season=max_season,
                season_length=season_length,
            )
        # Set the final one to be -1
        self.survey_features["N_obs_%i" % -1] = features.N_observations_season(
            -1,
            filtername=filtername,
            nside=self.nside,
            modulo=season_modulo,
            offset=day_offset,
            max_season=max_season,
            season_length=season_length,
        )
        self.survey_features["N_obs_count_all_%i" % -1] = features.N_obs_count_season(
            -1,
            filtername=None,
            season_modulo=season_modulo,
            offset=day_offset,
            nside=self.nside,
            max_season=max_season,
            season_length=season_length,
        )
        if target_maps is None:
            target_maps, labels = utils.get_current_footprint(nside)
            self.target_map = target_maps[filtername]
        else:
            self.target_maps = target_maps
        # should probably actually loop over all the target maps?
        self.out_of_bounds_area = np.where(self.target_maps[0] == 0)[0]
        self.out_of_bounds_val = out_of_bounds_val
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.all_indx = np.arange(self.result.size)

        # For computing what day each healpix is on
        if day_offset is None:
            self.day_offset = np.zeros(hp.nside2npix(self.nside), dtype=float)
        else:
            self.day_offset = day_offset

        self.season_modulo = season_modulo
        self.max_season = max_season
        self.season_length = season_length
        send_unused_deprecation_warning(self.__class__.__name__)

    def _calc_value(self, conditions, indx=None):
        """
        Parameters
        ----------
        indx : list (None)
            Index values to compute, if None, full map is computed
        Returns
        -------
        Healpix reward map
        """

        result = self.result.copy()
        if indx is None:
            indx = self.all_indx

        # Compute what season it is at each pixel
        seasons = utils.season_calc(
            conditions.night,
            offset=self.day_offset,
            modulo=self.season_modulo,
            max_season=self.max_season,
            season_length=self.season_length,
        )

        composite_target = self.result.copy()[indx]
        composite_nobs = self.result.copy()[indx]

        composite_goal_n = self.result.copy()[indx]

        for season in np.unique(seasons):
            season_indx = np.where(seasons == season)[0]
            composite_target[season_indx] = self.target_maps[season][season_indx]
            composite_nobs[season_indx] = self.survey_features["N_obs_%i" % season].feature[season_indx]
            composite_goal_n[season_indx] = (
                composite_target[season_indx]
                * self.survey_features["N_obs_count_all_%i" % season].feature
                * self.norm_factor
            )

        result[indx] = composite_goal_n - composite_nobs[indx]
        result[self.out_of_bounds_area] = self.out_of_bounds_val

        return result



# pylint: disable=too-many-arguments
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name

# imports
import unittest

import astropy
import numpy as np

from rubin_scheduler.utils import riseset

# constants

RANDOM_SEED = 6563
LOCATION = astropy.coordinates.EarthLocation.of_site("Cerro Pachon")


class TestRiseset(unittest.TestCase):
    num_times = 100
    tolerance = 1e-8
    alt_deg = -14.0

    def setUp(self):
        np.random.seed(RANDOM_SEED)
        self.request_mjds = 60000 + 365.24 * np.random.random(self.num_times)

    def test_riseset_times_down(self):
        for body in ("sun", "moon"):
            mjds = riseset.riseset_times(
                self.request_mjds,
                which_direction="down",
                which_night="nearest",
                alt=self.alt_deg,
                location=LOCATION,
                body=body,
                tolerance=self.tolerance,
            )
            self.assertTrue(np.allclose(mjds, self.request_mjds, atol=0.5))
            self.assertTrue(np.allclose(_body_alts(mjds, body), self.alt_deg, atol=self.tolerance))
            self.assertTrue(np.all(~_body_rising(mjds, body)))

    def test_riseset_times_up(self):
        for body in ("sun", "moon"):
            mjds = riseset.riseset_times(
                self.request_mjds,
                which_direction="up",
                which_night="nearest",
                alt=self.alt_deg,
                location=LOCATION,
                body=body,
                tolerance=self.tolerance,
            )
            self.assertTrue(np.allclose(mjds, self.request_mjds, atol=0.5))
            self.assertTrue(np.allclose(_body_alts(mjds, body), self.alt_deg, atol=self.tolerance))
            self.assertTrue(np.all(_body_rising(mjds, body)))

    def test_riseset_times_next_up(self):
        mjds = riseset.riseset_times(
            self.request_mjds,
            which_direction="up",
            which_night="next",
            alt=self.alt_deg,
            location=LOCATION,
            body="sun",
            tolerance=self.tolerance,
        )
        self.assertTrue(np.all(mjds >= self.request_mjds))
        self.assertTrue(np.allclose(_body_alts(mjds, "sun"), self.alt_deg, atol=self.tolerance))
        self.assertTrue(np.all(_body_rising(mjds, "sun")))

    def test_riseset_times_previous_up(self):
        mjds = riseset.riseset_times(
            self.request_mjds,
            which_direction="up",
            which_night="previous",
            alt=self.alt_deg,
            location=LOCATION,
            body="sun",
            tolerance=self.tolerance,
        )
        self.assertTrue(np.all(mjds <= self.request_mjds))
        self.assertTrue(np.allclose(_body_alts(mjds, "sun"), self.alt_deg, atol=self.tolerance))
        self.assertTrue(np.all(_body_rising(mjds, "sun")))

    def test_riseset_times_next_down(self):
        mjds = riseset.riseset_times(
            self.request_mjds,
            which_direction="down",
            which_night="next",
            alt=self.alt_deg,
            location=LOCATION,
            body="sun",
            tolerance=self.tolerance,
        )
        self.assertTrue(np.all(mjds >= self.request_mjds))
        self.assertTrue(np.allclose(_body_alts(mjds, "sun"), self.alt_deg, atol=self.tolerance))
        self.assertTrue(np.all(~_body_rising(mjds, "sun")))

    def test_riseset_times_previous_down(self):
        mjds = riseset.riseset_times(
            self.request_mjds,
            which_direction="down",
            which_night="previous",
            alt=self.alt_deg,
            location=LOCATION,
            body="sun",
            tolerance=self.tolerance,
        )
        self.assertTrue(np.all(mjds <= self.request_mjds))
        self.assertTrue(np.allclose(_body_alts(mjds, "sun"), self.alt_deg, atol=self.tolerance))
        self.assertTrue(np.all(~_body_rising(mjds, "sun")))


def _body_alts(mjds, body):
    times = astropy.time.Time(mjds, scale="utc", format="mjd", location=LOCATION)
    crds = astropy.coordinates.get_body(body, times, location=LOCATION)
    alt = crds.transform_to(astropy.coordinates.AltAz(obstime=times, location=LOCATION)).alt.deg
    return alt


def _body_rising(mjds, body):
    test_dt = 1 / (24 * 60.0)
    now_body_alts = _body_alts(mjds, body)
    future_body_alts = _body_alts(mjds + test_dt, body)
    rising = future_body_alts > now_body_alts
    return rising


run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



import unittest

import numpy as np
import pandas as pd

from rubin_scheduler.site_models import Almanac


class TestAlmanac(unittest.TestCase):
    def test_alm(self):
        alma = Almanac()

        mjd = 59853.35

        # Dead simple make sure the things load.
        planets = alma.get_planet_positions(mjd)
        sun = alma.get_sunset_info(mjd)
        moon = alma.get_sun_moon_positions(mjd)
        indx = alma.mjd_indx(mjd)

        assert planets is not None
        assert sun is not None
        assert moon is not None
        assert indx is not None

    def test_index_for_local_evening(self):
        almanac = Almanac()
        longitude = np.radians(-70.75)
        test_dates = ["2024-12-22", "2024-03-15", "2024-06-22"]
        for test_date in test_dates:
            night_index = almanac.index_for_local_evening(test_date, longitude)
            sunset_mjd = almanac.sunsets[night_index]["sunset"]
            sunset_timestamp = (sunset_mjd - 40587) * 24 * 60 * 60
            sunset_local_iso = (
                pd.Timestamp(sunset_timestamp, tz="UTC", unit="s").tz_convert("Chile/Continental").isoformat()
            )
            self.assertTrue(sunset_local_iso.startswith(test_date))


if __name__ == "__main__":
    unittest.main()



"""Tools for running the set of simulations used for a pre-night briefing."""

__all__ = [
    "AnomalousOverheadFunc",
    "run_prenights",
    "prenight_sim_cli",
]

import argparse
import io
import logging
import pickle
from datetime import datetime
from functools import partial
from tempfile import TemporaryFile
from typing import Callable, Optional, Sequence
from warnings import warn

import numpy as np
from astropy.time import Time
from matplotlib.pylab import Generator

from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers.core_scheduler import CoreScheduler
from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.sim_archive.sim_archive import drive_sim
from rubin_scheduler.site_models import Almanac

try:
    from rubin_sim.data import get_baseline  # type: ignore
except ModuleNotFoundError:
    get_baseline = partial(warn, "Cannot find default baseline because rubin_sim is not installed.")

DEFAULT_ARCHIVE_URI = "s3://rubin-scheduler-prenight/opsim/"


def _run_sim(
    mjd_start: float,
    archive_uri: str,
    scheduler_io: io.BufferedRandom,
    label: str,
    tags: Sequence[str] = tuple(),
    sim_length: float = 2,
    anomalous_overhead_func: Optional[Callable] = None,
) -> None:
    logging.info(f"Running {label}.")

    observatory = ModelObservatory(
        cloud_data="ideal",
        seeing_data="ideal",
        downtimes="ideal",
    )

    scheduler_io.seek(0)
    scheduler = pickle.load(scheduler_io)

    drive_sim(
        observatory=observatory,
        scheduler=scheduler,
        archive_uri=archive_uri,
        label=label,
        tags=tags,
        script=__file__,
        mjd_start=mjd_start,
        survey_length=sim_length,
        record_rewards=True,
        anomalous_overhead_func=anomalous_overhead_func,
    )


def _mjd_now() -> float:
    # Used instead of just Time.now().mjd to make type checker happy.
    mjd = Time.now().mjd
    assert isinstance(mjd, float)
    return mjd


def _iso8601_now() -> str:
    # Used instead of just Time.now().mjd to make type checker happy.
    now_iso = Time.now().iso
    assert isinstance(now_iso, str)
    return now_iso


def _create_scheduler_io(
    day_obs_mjd: float,
    scheduler_fname: Optional[str] = None,
    scheduler_instance: Optional[CoreScheduler] = None,
    opsim_db=None,
) -> io.BufferedRandom:
    if scheduler_instance is not None:
        scheduler: CoreScheduler = scheduler_instance
    elif scheduler_fname is None:
        sample_scheduler = example_scheduler()
        if not isinstance(sample_scheduler, CoreScheduler):
            raise TypeError()

        scheduler: CoreScheduler = sample_scheduler
    else:
        with open(scheduler_fname, "rb") as sched_io:
            scheduler = pickle.load(sched_io)

    scheduler.keep_rewards = True

    if opsim_db is not None:
        last_preexisting_obs_mjd = day_obs_mjd + 0.5
        obs = SchemaConverter().opsim2obs(opsim_db)
        obs = obs[obs["mjd"] < last_preexisting_obs_mjd]
        if len(obs) > 0:
            scheduler.add_observations_array(obs)

    scheduler_io = TemporaryFile()
    pickle.dump(scheduler, scheduler_io)
    scheduler_io.seek(0)
    return scheduler_io


class AnomalousOverheadFunc:
    """Callable to return random overhead.

    Parameters
    ----------
    seed : `int`
        Random number seed.
    slew_scale : `float`
        The scale for the scatter in the slew offest (seconds).
    visit_scale : `float`, optional
        The scale for the scatter in the visit overhead offset (seconds).
        Defaults to 0.0.
    slew_loc : `float`, optional
        The location of the scatter in the slew offest (seconds).
        Defaults to 0.0.
    visit_loc : `float`, optional
        The location of the scatter in the visit offset (seconds).
        Defaults to 0.0.
    """

    def __init__(
        self,
        seed: int,
        slew_scale: float,
        visit_scale: float = 0.0,
        slew_loc: float = 0.0,
        visit_loc: float = 0.0,
    ) -> None:
        self.rng: Generator = np.random.default_rng(seed)
        self.visit_loc: float = visit_loc
        self.visit_scale: float = visit_scale
        self.slew_loc: float = slew_loc
        self.slew_scale: float = slew_scale

    def __call__(self, visittime: float, slewtime: float) -> float:
        """Return a randomized offset for the visit overhead.

        Parameters
        ----------
        visittime : `float`
            The visit time (seconds).
        slewtime : `float`
            The slew time (seconds).

        Returns
        -------
        offset: `float`
            Random offset (seconds).
        """

        slew_overhead: float = slewtime * self.rng.normal(self.slew_loc, self.slew_scale)

        # Slew might be faster that expected, but it will never take negative
        # time.
        if (slewtime + slew_overhead) < 0:
            slew_overhead = 0.0

        visit_overhead: float = visittime * self.rng.normal(self.slew_loc, self.slew_scale)
        # There might be anomalous overhead that makes visits take longer,
        # but faster is unlikely.
        if visit_overhead < 0:
            visit_overhead = 0.0

        return slew_overhead + visit_overhead


def run_prenights(
    day_obs_mjd: float, archive_uri: str, scheduler_file: Optional[str] = None, opsim_db: Optional[str] = None
) -> None:
    """Run the set of scheduler simulations needed to prepare for a night.

    Parameters
    ----------
    day_obs_mjd : `float`
        The starting MJD>
    archive_uri : `str`
        The URI of visits completed before this night.
    scheduler_file : `str`, optional
        File from which to load the scheduler. None defaults to the example
        scheduler in rubin_sim, if it is installed.
        The default is None.
    opsim_db : `str`, optional
        The file name of the visit database for visits preceeding the
        simulation.
        The default is None.
    """

    exec_time: str = _iso8601_now()
    mjd_start: float = day_obs_mjd + 0.5
    scheduler_io = _create_scheduler_io(day_obs_mjd, scheduler_fname=scheduler_file, opsim_db=opsim_db)

    # Assign args common to all sims for this execution.
    run_sim = partial(_run_sim, archive_uri=archive_uri, scheduler_io=scheduler_io)

    # Begin with an ideal pure model sim
    run_sim(
        mjd_start,
        label=f"Nominal start and overhead, ideal conditions, run at {exec_time}",
        tags=["ideal", "nominal"],
    )

    # Delayed start
    # Almanac.get_sunset_info does not use day_obs, so just index
    # Almanac.sunsets for what we want directly.
    all_sun_n12_setting = Almanac().sunsets["sun_n12_setting"]
    before_day_obs = all_sun_n12_setting < day_obs_mjd + 0.5
    after_day_obs = all_sun_n12_setting > day_obs_mjd + 1.5
    on_day_obs = ~(before_day_obs | after_day_obs)
    sun_n12_setting = all_sun_n12_setting[on_day_obs].item()

    for minutes_delay in (0, 1, 10, 60, 240):
        delayed_mjd_start = sun_n12_setting + minutes_delay / (24.0 * 60)

        # 2 nights of observing, 0.5 nights to shift to the day_obs rollover
        sim_length = day_obs_mjd + 2.5 - delayed_mjd_start

        run_sim(
            delayed_mjd_start,
            label=f"Start time delayed by {minutes_delay} minutes,"
            + f" Nominal slew and visit overhead, ideal conditions, run at {exec_time}",
            tags=["ideal", f"delay_{minutes_delay}"],
            sim_length=sim_length,
        )

    # Run a few different scatters of visit time
    anomalous_overhead_scale = 0.1
    for anomalous_overhead_seed in range(101, 106):
        anomalous_overhead_func = AnomalousOverheadFunc(anomalous_overhead_seed, anomalous_overhead_scale)
        run_sim(
            mjd_start,
            label=f"Anomalous overhead {anomalous_overhead_seed, anomalous_overhead_scale},"
            + f" Nominal start, ideal conditions, run at {exec_time}",
            tags=["ideal", "anomalous_overhead"],
            anomalous_overhead_func=anomalous_overhead_func,
            sim_length=2,
        )


def _parse_dayobs_to_mjd(dayobs: str | float) -> float:
    try:
        day_obs_mjd = Time(dayobs).mjd
    except ValueError:
        try:
            day_obs_mjd = Time(datetime.strptime(str(dayobs), "%Y%m%d"))
        except ValueError:
            day_obs_mjd = Time(dayobs, format="mjd")

    if not isinstance(day_obs_mjd, float):
        raise ValueError

    return day_obs_mjd


def prenight_sim_cli(*args):
    if Time.now() >= Time("2025-05-01"):
        default_time = Time(int(_mjd_now() - 0.5), format="mjd")
    else:
        default_time = Time("2025-05-14")

    parser = argparse.ArgumentParser(description="Run prenight simulations")
    parser.add_argument(
        "--dayobs",
        type=str,
        default=default_time.iso,
        help="The day_obs of the night to simulate.",
    )
    parser.add_argument(
        "--archive",
        type=str,
        default=DEFAULT_ARCHIVE_URI,
        help="Archive in which to store simulation results.",
    )
    parser.add_argument("--scheduler", type=str, default=None, help="pickle file of the scheduler to run.")

    # Only pass a default if we have an opsim
    opsim_arg_kwargs = {"type": str, "help": "Opsim database from which to load previous visits."}
    baseline = get_baseline()
    if baseline is not None:
        opsim_arg_kwargs["default"] = baseline

    parser.add_argument("--opsim", **opsim_arg_kwargs)
    args = parser.parse_args() if len(args) == 0 else parser.parse_args(args)

    day_obs_mjd = _parse_dayobs_to_mjd(args.dayobs)
    archive_uri = args.archive
    scheduler_file = args.scheduler
    opsim_db = args.opsim

    run_prenights(day_obs_mjd, archive_uri, scheduler_file, opsim_db)


if __name__ == "__main__":
    prenight_sim_cli()



__all__ = ("Conditions",)

import functools
import warnings
from io import StringIO

import healpy as hp
import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.utils import match_hp_resolution, set_default_nside, smallest_signed_angle
from rubin_scheduler.utils import (
    Site,
    _angular_separation,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    _hpid2_ra_dec,
    calc_lmst,
    calc_season,
    m5_flat_sed,
    survey_start_mjd,
)


class Conditions:
    """Holds telemetry information, keeping calculated values in sync
    for `self.mjd` (such as ra/dec mappings to alt/az).

    Incoming values have setters to keep values in sync. Healpix maps
    are set to the expected (`self.nside`) resolution.

    Unless otherwise noted, all values are assumed to be valid at the time
    given by `self.mjd`.

    Parameters
    ----------
    nside : `int`, optional
        The healpixel nside to set the resolution of attributes.
        Default of None will use
        `rubin_scheduler.scheduler.utils.set_default_nside`.
    site : str ('LSST')
        A site name used to create a sims.utils.Site object. For
        looking up observatory parameters like latitude and longitude.
    exptime : `float`, optional
        The exposure time (seconds) to assume when computing the 5-sigma
        limiting depth maps stored in Conditions. These maps are used
        by basis functions, but are not used to calculate expected
        observation m5 values (such as when exposure time varies).
        Default 30 seconds.
    mjd_start : `float`, optional
        The starting MJD of the survey. Default uses
        `rubin_scheduler.utils.survey_start_date()`.
    season_map : `np.array`, (N,), optional
        A HEALpix array that specifies the day offset when computing
        the season for each HEALpix. Equivalent to season at mjd_start.
    sun_ra_start : `float`, optional
        The RA of the sun at the start of the survey (radians)
    mjd : `float`, optional
        The current MJD.
        Default of None is fine on init - will be updated by telemetry
        stream.
    """

    global_maps = set(["ra", "dec", "slewtime", "airmass", "zeros_map", "nan_map"])
    by_band_maps = set(["skybrightness", "fwhm_eff", "m5_depth"])

    def __init__(
        self,
        nside=None,
        site="LSST",
        exptime=30.0,
        mjd_start=survey_start_mjd(),
        season_map=None,
        sun_ra_start=None,
        mjd=None,
    ):
        """
        Attributes (Set on init)
        -----------
        nside : int
            Healpix resolution. All maps are set to this reslution.
        site : rubin_scheduler.Site object ('LSST')
            Contains static site-specific data (lat, lon, altitude, etc).
            Defaults to 'LSST'.
        ra : np.array
            A healpix array with the RA of each healpixel center (radians).
            Automatically generated.
        dec : np.array
            A healpix array with the Dec of each healpixel center (radians).
            Automatically generated.

        Attributes (to be set by user/telemetry stream/scheduler)
        -------------------------------------------
        mjd : float
            Modified Julian Date (days).
        bulk_cloud : float
            The fraction of sky covered by clouds. (In the future might
            update to transparency map)
        cloud_map : np.array
            XXX--to be done. HEALpix array with cloudy pixels set to NaN.
        slewtime : np.array
            Healpix showing the slewtime to each healpixel center (seconds)
        current_filter : str
            The name of the current filter. (expect one of u, g, r, i, z, y).
        mounted_filters : list of str
            The filters that are currently mounted and thu available
            (expect 5 of u, g, r, i, z, y)
        night : int
            The current night number (days). Probably starts at 1.
        skybrightness : dict of np.array
            Dictionary keyed by filtername. Values are healpix arrays with
            the sky brightness at each
            healpix center (mag/acsec^2)
        fwhm_eff : dict of np.array
            Dictionary keyed by filtername. Values are the effective seeing
            FWHM at each healpix
            center (arcseconds)
        moon_alt : float
            The altitude of the Moon (radians)
        moon_az : float
            The Azimuth of the moon (radians)
        moon_ra : float
            RA of the moon (radians)
        moon_dec : float
            Declination of the moon (radians)
        moon_phase : float
            The Phase of the moon. (percent, 0=new moon, 100=full moon)
        sun_alt : float
            The altitude of the sun (radians).
        sun_az : float
            The Azimuth of the sun (radians).
        sun_ra : float
            The RA of the sun (radians).
        sun_dec : float
            The Dec of the sun (radians).
        tel_ra : float
            The current telescope RA pointing (radians).
        tel_dec : float
            The current telescope Declination (radians).
        tel_alt : float
            The current telescope altitude (radians).
        tel_az : float
            The current telescope azimuth (radians).
        cumulative_azimuth_rad : float
            The cummulative telescope azimuth (radians). For tracking
            cable wrap
        cloud_map : np.array
            A healpix map with the cloud coverage. XXX-expand,
            is this bool map? Transparency map?
        airmass : np.array
            A healpix map with the airmass value of each healpixel. (unitless)
        wind_speed : float
            Wind speed (m/s).
        wind_direction : float
            Direction from which the wind originates. A direction of
            0.0 degrees means the wind originates from the north and 90.0
            degrees from the east (radians).
        sunset : float
            The MJD of sunset that starts the current night. Note MJDs of
            sunset, moonset, twilight times, etc are from interpolations.
            This means the sun may actually be slightly above/below the
            horizon at the given sunset time.
        sun_n12_setting : float
            The MJD of when the sun is at -12 degees altitude and
            setting during the current night. From interpolation.
        sun_n18_setting : float
            The MJD when the sun is at -18 degrees altitude and setting
            during the current night. From interpolation.
        sun_n18_rising : float
            The MJD when the sun is at -18 degrees altitude and rising
            during the current night. From interpolation.
        sun_n12_rising : float
            The MJD when the sun is at -12 degrees altitude and rising
            during the current night. From interpolation.
        sunrise : float
            The MJD of sunrise during the current night. From interpolation
        mjd_start : float
            The starting MJD of the survey.
        moonrise : float
            The MJD of moonrise during the current night. From interpolation.
        moonset : float
            The MJD of moonset during the current night. From interpolation.
        moon_phase_sunset : float
            The phase of the moon (0-100 illuminater) at sunset.
            Useful for setting which filters should be loaded.
        targets_of_opportunity : list of rubin_scheduler.scheduler.targetoO
            targetoO objects.
        planet_positions : dict
            Dictionary of planet name and coordinate e.g., 'venus_RA',
            'mars_dec'
        scheduled_observations : np.array
            A list of MJD times when there are scheduled observations.
            Defaults to empty array.
        tel_az_limits : list of float pairs
            A list of lists giving valid azimuth ranges. e.g.,
            [0, 2*np.pi] would mean all azimuth values are valid, while
            [[0, np.pi/2], [3*np.pi/2, 2*np.pi]] would mean anywhere in
            the south is invalid.  Radians.
        tel_alt_limits : list of float pairs
            A list of lists giving valid altitude ranges. Radians.

        Attributes (calculated on demand and cached)
        ------------------------------------------
        alt : np.array
            Altitude of each healpixel (radians). Recaclulated if mjd is
            updated. Uses fast approximate equation for converting
            RA,Dec to alt,az.
        az : np.array
            Azimuth of each healpixel (radians). Recaclulated if mjd is
            updated. Uses fast approximate equation for converting RA,Dec
            to alt,az.
        pa : np.array
            The parallactic angle of each healpixel (radians). Recaclulated
            if mjd is updated. Based on the fast approximate alt,az values.
        season : `np.array`, (N,)
            The current season value (0-1) at each healpixel.
            Recalculated if mjd is updated. Used by features that would
            otherwise need to calculate `season`.
            Uses `rubin_scheduler.utils.calc_season`.
        lmst : float
            The local mean sidearal time (hours). Updates is mjd is changed.
        m5_depth : dict of np.array
            the 5-sigma limiting depth healpix maps, keyed by filtername
            (mags). Will be recalculated if the skybrightness, seeing,
            or airmass are updated.
        HA : np.array
            Healpix map of the hour angle of each healpixel (radians).
            Runs from 0 to 2pi.
        az_to_sun : np.array
            Healpix map of the azimuthal distance to the sun for each
            healpixel (radians)
        az_to_anitsun : np.array
            Healpix map of the azimuthal distance to the anit-sun for each
            healpixel (radians)
        solar_elongation : np.array
            Healpix map of the solar elongation (angular distance to the sun)
            for each healpixel (radians)

        Attributes (set by the scheduler)
        -------------------------------
        queue : list of observation objects
            The current queue of observations core_scheduler is waiting to
            execute.

        """
        if nside is None:
            nside = set_default_nside()
        self.nside = nside

        # The RA, Dec grid we are using
        hpids = np.arange(hp.nside2npix(self.nside))
        self.ra, self.dec = _hpid2_ra_dec(self.nside, hpids)

        self.site = Site(site)
        self.exptime = exptime

        self.mjd_start = mjd_start

        # Keep these named in both ways, for backwards compatibility for now
        if season_map is None:
            season_map = calc_season(np.degrees(self.ra), [self.mjd_start], self.mjd_start).flatten()
        # season_offset in units of days -useful for season_calc calculations
        self.season_offset = season_map * 365.25
        # Season_map is useful for utils.calc_season calculations
        self.season_map = season_map

        self.sun_ra_start = sun_ra_start

        # Generate an empty map so we can copy when we need a new map
        self.zeros_map = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.nan_map = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.nan_map.fill(np.nan)

        # Set other attributes to Nones
        self._init_attributes()
        self.mjd = mjd

    def _init_attributes(self):
        """Initialize or clear all the attributes"""

        # Modified Julian Date (day)
        self._mjd = None
        # Altitude and azimuth. Dict with degrees and radians
        self._alt = None
        self._az = None
        self._pa = None

        # The cloud level. Fraction, but could upgrade to transparency map
        self.clouds = None
        self._slewtime = None
        self.current_filter = None
        self.mounted_filters = None
        self.night = None
        self._lmst = None
        # Should be a dict with filtername keys
        self._skybrightness = {}
        self._fwhm_eff = {}
        self._m5_depth = None
        self._airmass = None

        self.wind_speed = None
        self.wind_direction = None

        # Upcoming scheduled observations
        self.scheduled_observations = np.array([], dtype=float)

        # Attribute to hold the current observing queue
        self.queue = None

        # Moon
        self.moon_alt = None
        self.moon_az = None
        self.moon_ra = None
        self.moon_dec = None
        self.moon_phase = None

        # Sun
        self.sun_alt = None
        self.sun_az = None
        self.sun_ra = None
        self.sun_dec = None

        # Almanac information
        self.sunset = None
        self.sun_n12_setting = None
        self.sun_n18_setting = None
        self.sun_n18_rising = None
        self.sun_n12_rising = None
        self.sunrise = None
        self.moonrise = None
        self.moonset = None
        self.moon_phase_sunset = None

        self.planet_positions = None

        # Current telescope pointing
        self.tel_ra = None
        self.tel_dec = None
        self.tel_alt = None
        self.tel_az = None
        self.cumulative_azimuth_rad = None

        # Telescope limits
        self.tel_az_limits = None
        self.tel_alt_limits = None

        # Full sky cloud map
        self._cloud_map = None
        self._HA = None

        # XXX--document
        self.bulk_cloud = None

        self.rot_tel_pos = None

        self.targets_of_opportunity = None

        # self._season tracks the actual current
        # observing season at each pixel
        self._season = None
        # int_season = np.floor(season)
        self._int_season = None

        # Potential attributes that get computed
        self._solar_elongation = None
        self._az_to_sun = None
        self._az_to_antisun = None

    @property
    def lmst(self):
        if self._lmst is None:
            self._lmst = calc_lmst(self.mjd, self.site.longitude_rad)

        return self._lmst

    @lmst.setter
    def lmst(self, value):
        self._lmst = value
        self._HA = None

    @property
    def HA(self):
        if self._HA is None:
            self.calc_ha()
        return self._HA

    def calc_ha(self):
        self._HA = np.radians(self.lmst * 360.0 / 24.0) - self.ra
        self._HA[np.where(self._HA < 0)] += 2.0 * np.pi

    @property
    def cloud_map(self):
        return self._cloud_map

    @cloud_map.setter
    def cloud_map(self, value):
        self._cloud_map = match_hp_resolution(value, nside_out=self.nside)

    @property
    def slewtime(self):
        return self._slewtime

    @slewtime.setter
    def slewtime(self, value):
        # Using 0 for start of night
        if np.size(value) == 1:
            self._slewtime = value
        else:
            self._slewtime = match_hp_resolution(value, nside_out=self.nside)

    @property
    def airmass(self):
        return self._airmass

    @airmass.setter
    def airmass(self, value):
        self._airmass = match_hp_resolution(value, nside_out=self.nside)
        self._m5_depth = None

    @property
    def pa(self):
        if self._pa is None:
            self.calc_pa()
        return self._pa

    def calc_pa(self):
        self._pa = _approx_altaz2pa(self.alt, self.az, self.site.latitude_rad)

    @property
    def alt(self):
        if self._alt is None:
            self.calc_alt_az()
        return self._alt

    @property
    def az(self):
        if self._az is None:
            self.calc_alt_az()
        return self._az

    def calc_alt_az(self):
        self._alt, self._az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            self._mjd,
        )

    @property
    def season(self):
        if self._season is None:
            self.update_season()
        return self._season

    @property
    def int_season(self):
        if self._int_season is None:
            self.update_season()
        return self._int_season

    def update_season(self):
        self._season = self.season_map + (self.mjd - self.mjd_start) / 365.25
        self._int_season = np.floor(self._season)

    @functools.lru_cache(maxsize=10)
    def future_alt_az(self, mjd):
        """Compute the altitude and azimuth for a future time.

        Returns
        -------
        altitude : `np.array`
            The altutude of each healpix at MJD (radians)
        azimuth : : `np.array`
            The azimuth of each healpix at MJD (radians)
        """
        alt, az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            mjd,
        )
        return alt, az

    @property
    def mjd(self):
        return self._mjd

    @mjd.setter
    def mjd(self, value):
        # If MJD is changed, everything else is no longer valid, so re-init
        if self.mjd is not None:
            warnings.warn("Changing MJD and resetting all attributes.")
        self._init_attributes()
        self._mjd = value

    @property
    def skybrightness(self):
        return self._skybrightness

    @skybrightness.setter
    def skybrightness(self, indict):
        for key in indict:
            self._skybrightness[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        # If sky brightness changes, need to recalc M5 depth.
        self._m5_depth = None

    @property
    def fwhm_eff(self):
        return self._fwhm_eff

    @fwhm_eff.setter
    def fwhm_eff(self, indict):
        for key in indict:
            self._fwhm_eff[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        self._m5_depth = None

    @property
    def m5_depth(self):
        if self._m5_depth is None:
            self.calc_m5_depth()
        return self._m5_depth

    def calc_m5_depth(self):
        self._m5_depth = {}
        for filtername in self._skybrightness:
            good = ~np.isnan(self._skybrightness[filtername])
            self._m5_depth[filtername] = self.nan_map.copy()
            self._m5_depth[filtername][good] = m5_flat_sed(
                filtername,
                self._skybrightness[filtername][good],
                self._fwhm_eff[filtername][good],
                self.exptime,
                self._airmass[good],
            )

    def calc_solar_elongation(self):
        self._solar_elongation = _angular_separation(self.ra, self.dec, self.sun_ra, self.sun_dec)

    @property
    def solar_elongation(self):
        if self._solar_elongation is None:
            self.calc_solar_elongation()
        return self._solar_elongation

    def calc_az_to_sun(self):
        self._az_to_sun = smallest_signed_angle(self.ra, self.sun_ra)

    def calc_az_to_antisun(self):
        self._az_to_antisun = smallest_signed_angle(self.ra + np.pi, self.sun_ra)

    @property
    def az_to_sun(self):
        if self._az_to_sun is None:
            self.calc_az_to_sun()
        return self._az_to_sun

    @property
    def az_to_antisun(self):
        if self._az_to_antisun is None:
            self.calc_az_to_antisun()
        return self._az_to_antisun

    def __repr__(self):
        return f"<{self.__class__.__name__} mjd_start='{self.mjd_start}' at {hex(id(self))}>"

    def __str__(self):
        # If dependencies of to_markdown are not installed, fall back on repr
        try:
            pd.DataFrame().to_markdown()
        except ImportError:
            return repr(self)

        if self.mjd is None:
            # Many elements of the pretty str representation rely on
            # the time beging set. If it is not, just return the ugly form.
            return repr(self)

        output = StringIO()
        print(f"{self.__class__.__qualname__} at {hex(id(self))}", file=output)
        print("============================", file=output)
        print("nside: ", self.nside, "  ", file=output)
        print("site: ", self.site.name, "  ", file=output)
        print("exptime: ", self.exptime, "  ", file=output)
        print("lmst: ", self.lmst, "  ", file=output)
        print("season_offset: ", self.season_offset, "  ", file=output)
        print("sun_RA_start: ", self.sun_ra_start, "  ", file=output)
        print("clouds: ", self.clouds, "  ", file=output)
        print("current_filter: ", self.current_filter, "  ", file=output)
        print("mounted_filters: ", self.mounted_filters, "  ", file=output)
        print("night: ", self.night, "  ", file=output)
        print("wind_speed: ", self.wind_speed, "  ", file=output)
        print("wind_direction: ", self.wind_direction, "  ", file=output)
        print(
            "len(scheduled_observations): ",
            len(self.scheduled_observations),
            "  ",
            file=output,
        )
        print(
            "len(queue): ",
            None if self.queue is None else len(self.queue),
            "  ",
            file=output,
        )
        print("moonPhase: ", self.moon_phase, "  ", file=output)
        print("bulk_cloud: ", self.bulk_cloud, "  ", file=output)
        print("targets_of_opportunity: ", self.targets_of_opportunity, "  ", file=output)
        print("cumulative_azimuth_rad: ", self.cumulative_azimuth_rad, "  ", file=output)

        positions = [
            {
                "name": "sun",
                "alt": self.sun_alt,
                "az": self.sun_az,
                "RA": self.sun_ra,
                "decl": self.sun_dec,
            }
        ]
        positions.append(
            {
                "name": "moon",
                "alt": self.moon_alt,
                "az": self.moon_az,
                "RA": self.moon_ra,
                "decl": self.moon_dec,
            }
        )
        for planet_name in ("venus", "mars", "jupiter", "saturn"):
            positions.append(
                {
                    "name": planet_name,
                    "RA": float(self.planet_positions[planet_name + "_RA"]),
                    "decl": float(self.planet_positions[planet_name + "_dec"]),
                }
            )
        positions.append(
            {
                "name": "telescope",
                "alt": self.tel_alt,
                "az": self.tel_az,
                "RA": self.tel_ra,
                "decl": self.tel_dec,
                "rot": self.rot_tel_pos,
            }
        )
        positions = pd.DataFrame(positions).set_index("name")
        print(file=output)
        print("Positions (radians)", file=output)
        print("-------------------", file=output)
        print(positions.to_markdown(), file=output)

        positions_deg = np.degrees(positions)
        print(file=output)
        print("Positions (degrees)", file=output)
        print("-------------------", file=output)
        print(positions_deg.to_markdown(), file=output)

        events = (
            "mjd_start",
            "mjd",
            "sunset",
            "sun_n12_setting",
            "sun_n18_setting",
            "sun_n18_rising",
            "sun_n12_rising",
            "sunrise",
            "moonrise",
            "moonset",
            "sun_0_setting",
            "sun_0_rising",
        )
        event_rows = []
        for event in events:
            try:
                mjd = getattr(self, event)
                time = pd.to_datetime(float(mjd) + 2400000.5, unit="D", origin="julian")
                event_rows.append({"event": event, "MJD": mjd, "date": time})
            except AttributeError:
                pass

        event_df = pd.DataFrame(event_rows).set_index("event").sort_values(by="MJD")
        print("", file=output)
        print("Events", file=output)
        print("------", file=output)
        print(event_df.to_markdown(), file=output)

        map_stats = []
        for map_name in self.global_maps - set(["zeros_map", "nan_map"]):
            try:
                values = getattr(self, map_name)
                map_stats.append(
                    {
                        "map": map_name,
                        "nside": hp.npix2nside(len(values)),
                        "min": np.nanmin(values),
                        "max": np.nanmax(values),
                        "median": np.nanmedian(values),
                    }
                )
            except AttributeError:
                pass

        for base_map_name in self.by_band_maps:
            for band in "ugrizy":
                try:
                    values = getattr(self, base_map_name)[band]
                    map_name = f"{base_map_name}_{band}"
                    map_stats.append(
                        {
                            "map": map_name,
                            "nside": hp.npix2nside(len(values)),
                            "min": np.nanmin(values),
                            "max": np.nanmax(values),
                            "median": np.nanmedian(values),
                        }
                    )
                except AttributeError:
                    pass

        maps_df = pd.DataFrame(map_stats).set_index("map")
        print("", file=output)
        print("Maps", file=output)
        print("----", file=output)
        print(maps_df.to_markdown(), file=output)

        result = output.getvalue()
        return result

    def _repr_markdown_(self):
        return str(self)



import importlib.util
import unittest
from tempfile import TemporaryDirectory

from rubin_scheduler.sim_archive import prenight_sim_cli

try:
    from lsst.resources import ResourcePath

    HAVE_RESOURCES = True
except ModuleNotFoundError:
    HAVE_RESOURCES = False

# We need rubin_sim to get the baseline sim
# Tooling prefers checking that it exists using importlib rather
# than importing it and not actually using it.
HAVE_RUBIN_SIM = importlib.util.find_spec("rubin_sim")


class TestPrenight(unittest.TestCase):
    @unittest.skip("Too slow")
    @unittest.skipIf(not HAVE_RESOURCES, "No lsst.resources")
    @unittest.skipIf(not HAVE_RUBIN_SIM, "No rubin_sim, needed for rubin_sim.data.get_baseline")
    def test_prenight(self):
        with TemporaryDirectory() as test_archive_dir:
            archive_uri = ResourcePath(test_archive_dir).geturl()  # type: ignore
            prenight_sim_cli("--archive", archive_uri)



from .base_survey import *
from .dd_surveys import *
from .ddf_presched import *
from .desc_ddf import *
from .field_survey import *
from .generate_ddf_grid import *
from .long_gap_survey import *
from .plan_night_survey import *
from .pointings_survey import *
from .scripted_surveys import *
from .surveys import *
from .too_surveys import *



__all__ = ("LsstCameraFootprint",)

import os

import numpy as np

from rubin_scheduler.data import get_data_dir

from .projections import gnomonic_project_toxy


def rotate(x, y, rotation_angle_rad):
    """rotate 2d points around the origin (0,0)"""
    cos_rad = np.cos(rotation_angle_rad)
    sin_rad = np.sin(rotation_angle_rad)
    qx = cos_rad * x - sin_rad * y
    qy = sin_rad * x + cos_rad * y
    return qx, qy


class LsstCameraFootprint:
    """
    Identify point on the sky within an LSST camera footprint.

    Parameters
    ----------
    units : `str`, opt
        Units for the object RA/Dec and boresight RA/Dec/rotation values.
        Default 'degrees'.  If not degrees, assumes incoming values
        are in radians.
    footprint_file : `str` or None, opt
        Location for the camera footprint map.
        Default None loads the default map from
        $RUBIN_SIM_DATA_DIR/maf/fov_map.npz

    """

    def __init__(self, units="degrees", footprint_file=None):
        if footprint_file is None:
            footprint_file = os.path.join(get_data_dir(), "utils", "fov_map.npz")
        _temp = np.load(footprint_file)
        # Units refers to the incoming ra/dec values in the _call__ method
        # Internally, radians are used to calculate the footprint
        self.units = units
        self.camera_fov = _temp["image"].copy()
        self.x_camera = np.radians(_temp["x"].copy())
        _temp.close()
        self.plate_scale = self.x_camera[1] - self.x_camera[0]
        self.indx_max = len(self.x_camera)

    def __call__(self, obj_ra, obj_dec, boresight_ra, boresight_dec, boresight_rot_sky_pos):
        """Determine which observations are within the actual camera
        footprint for a series of observations.

        Parameters
        ----------
        obj_ra : `np.ndarray`
            RA values for the objects (in the field of view?)
        obj_dec : `np.ndarray`
            Dec values for the objects
        boresight_ra : `float`
            RA value for the pointing
        boresight_dec : `float`
            Dec value for the pointing
        boresight_rot_sky_pos : `float`
            RotSkyPos value for the pointing.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object
            observations which are inside the fov and land on a
            science chip. Applying this to the input array
            (e.g. obj_ra[indices]) indicates the positions of
            the objects which fell onto active silicon.
        """
        if self.units == "degrees":
            x_proj, y_proj = gnomonic_project_toxy(
                np.radians(obj_ra),
                np.radians(obj_dec),
                np.radians(boresight_ra),
                np.radians(boresight_dec),
            )
            # rotate them by rotskypos
            # TODO: look up whether this is a positive or
            # negative rotation in the observatory documentation
            x_proj, y_proj = rotate(x_proj, y_proj, np.radians(boresight_rot_sky_pos))

        else:
            x_proj, y_proj = gnomonic_project_toxy(obj_ra, obj_dec, boresight_ra, boresight_dec)
            x_proj, y_proj = rotate(x_proj, y_proj, boresight_rot_sky_pos)

        # look up which points are good
        x_indx = np.round((x_proj - self.x_camera[0]) / self.plate_scale).astype(int)
        y_indx = np.round((y_proj - self.x_camera[0]) / self.plate_scale).astype(int)
        in_range = np.where(
            (x_indx >= 0) & (x_indx < self.indx_max) & (y_indx >= 0) & (y_indx < self.indx_max)
        )[0]
        x_indx = x_indx[in_range]
        y_indx = y_indx[in_range]
        indices = in_range
        # reduce the indices down to only the ones that fall on silicon.
        # self.camera_fov is an array of `bool` values
        map_val = self.camera_fov[x_indx, y_indx]
        indices = indices[map_val]
        return indices



import unittest

import healpy as hp
import numpy as np

import rubin_scheduler.utils as utils


class TestHealUtils(unittest.TestCase):
    def test_ra_decs_rad(self):
        """
        Test that the Ra Dec conversions round-trip
        """

        nside = 64
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils._hpid2_ra_dec(nside, hpids)

        hpids_return = utils._ra_dec2_hpid(nside, ra, dec)

        np.testing.assert_array_equal(hpids, hpids_return)

    def test_ra_decs_deg(self):
        """
        Test that the Ra Dec conversions round-trip
        """

        nside = 64
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils.hpid2_ra_dec(nside, hpids)

        hpids_return = utils.ra_dec2_hpid(nside, ra, dec)

        np.testing.assert_array_equal(hpids, hpids_return)

    def test_bin_rad(self):
        """
        Test that healbin returns correct values and valid healpy maps.
        """

        ra = np.zeros(3)
        dec = np.zeros(3)
        values = ra * 0.0 + 1.0

        nside = 128
        hpid = utils._ra_dec2_hpid(nside, ra[0], dec[0])

        map1 = utils._healbin(ra, dec, values, nside=nside)
        self.assertEqual(map1[hpid], 1.0)
        self.assertEqual(hp.maptype(map1), 0)
        map2 = utils._healbin(ra, dec, values, nside=nside, reduce_func=np.sum)
        self.assertEqual(map2[hpid], 3.0)
        self.assertEqual(hp.maptype(map2), 0)
        map3 = utils._healbin(ra, dec, values, nside=nside, reduce_func=np.std)
        self.assertEqual(map3[hpid], 0.0)
        self.assertEqual(hp.maptype(map3), 0)

    def test_bin_deg(self):
        """
        Test that healbin returns correct values and valid healpy maps.
        """

        ra = np.zeros(3)
        dec = np.zeros(3)
        values = ra * 0.0 + 1.0

        nside = 128
        hpid = utils.ra_dec2_hpid(nside, ra[0], dec[0])

        map1 = utils.healbin(ra, dec, values, nside=nside)
        self.assertEqual(map1[hpid], 1.0)
        self.assertEqual(hp.maptype(map1), 0)
        map2 = utils.healbin(ra, dec, values, nside=nside, reduce_func=np.sum)
        self.assertEqual(map2[hpid], 3.0)
        self.assertEqual(hp.maptype(map2), 0)
        map3 = utils.healbin(ra, dec, values, nside=nside, reduce_func=np.std)
        self.assertEqual(map3[hpid], 0.0)
        self.assertEqual(hp.maptype(map3), 0)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import rubin_scheduler.scheduler.basis_functions as basis_functions
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.utils import empty_observation


class TestBasis(unittest.TestCase):
    def test_visit_repeat_basis_function(self):
        bf = basis_functions.VisitRepeatBasisFunction()

        indx = np.array([1000])

        # 30 minute step
        delta = 30.0 / 60.0 / 24.0

        # Add 1st observation, should still be zero
        obs = empty_observation()
        obs["filter"] = "r"
        obs["mjd"] = 59000.0
        conditions = Conditions()
        conditions.mjd = np.max(obs["mjd"])
        bf.add_observation(obs, indx=indx)
        self.assertEqual(np.max(bf(conditions)), 0.0)

        # Advance time so now we want a pair
        conditions.mjd += delta
        self.assertEqual(np.max(bf(conditions)), 1.0)

        # Now complete the pair and it should go back to zero
        bf.add_observation(obs, indx=indx)

        conditions.mjd += delta
        self.assertEqual(np.max(bf(conditions)), 0.0)

    def test_label(self):
        bf = basis_functions.VisitRepeatBasisFunction()
        self.assertIsInstance(bf.label(), str)

        bf = basis_functions.SlewtimeBasisFunction(nside=16)
        self.assertIsInstance(bf.label(), str)

    def test_visit_gap(self):
        visit_gap = basis_functions.VisitGap(note="test")

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = empty_observation()
        observation["filter"] = "r"
        observation["note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["note"] = "test"
        visit_gap.add_observation(observation=observation)

        # now observation with the correct note
        assert not visit_gap.check_feasibility(conditions=conditions)

        # check it becomes feasible again once enough time has passed
        conditions.mjd += 2.0 * visit_gap.gap

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_visit_gap_with_filter(self):
        visit_gap = basis_functions.VisitGap(note="test", filter_names=["g"])

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = empty_observation()
        observation["filter"] = "r"
        observation["note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["note"] = "test"
        visit_gap.add_observation(observation=observation)

        # observation with the wrong filter
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["filter"] = "g"
        visit_gap.add_observation(observation=observation)

        # observation with the correct note and filter
        assert not visit_gap.check_feasibility(conditions=conditions)

        # check it becomes feasible again once enough time has passed
        conditions.mjd += 2.0 * visit_gap.gap

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_visit_gap_with_multiple_filters(self):
        visit_gap = basis_functions.VisitGap(note="test", filter_names=["g", "i"])

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = empty_observation()
        observation["filter"] = "r"
        observation["note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["note"] = "test"
        visit_gap.add_observation(observation=observation)

        # observation with the wrong filter
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["filter"] = "g"
        observation["mjd"] += 1e-3
        visit_gap.add_observation(observation=observation)

        # observation with the correct note but only one filter
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["filter"] = "i"
        observation["mjd"] += 1e-3
        visit_gap.add_observation(observation=observation)

        # observation with the correct note and both filters
        assert not visit_gap.check_feasibility(conditions=conditions)

        # make sure it is still not feasible after only the g observation gap
        # has passed
        conditions.mjd += visit_gap.gap + 1.1e-3

        # observation with the correct note and both filters
        assert not visit_gap.check_feasibility(conditions=conditions)

        # make sure it is feasible after both gaps have passed
        conditions.mjd += 1e-3

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_sun_alt(self):
        sunaltbf = basis_functions.SunAltHighLimitBasisFunction(alt_limit=-15)
        conditions = Conditions()
        conditions.sun_alt = np.radians(-20)
        assert ~sunaltbf.check_feasibility(conditions)
        conditions.sun_alt = np.radians(-10)
        assert sunaltbf.check_feasibility(conditions)


if __name__ == "__main__":
    unittest.main()



import unittest
from copy import deepcopy

import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import HpInLsstFov, empty_observation
from rubin_scheduler.skybrightness_pre import dark_sky
from rubin_scheduler.utils import survey_start_mjd


class TestFeatures(unittest.TestCase):
    def test_pair_in_night(self):
        pin = features.PairInNight(gap_min=25.0, gap_max=45.0)
        self.assertEqual(np.max(pin.feature), 0.0)

        indx = np.array([1000])

        delta = 30.0 / 60.0 / 24.0

        # Add 1st observation, feature should still be zero
        obs = empty_observation()
        obs["filter"] = "r"
        obs["mjd"] = 59000.0
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 0.0)

        # Add 2nd observation
        obs["mjd"] += delta
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 1.0)

        obs["mjd"] += delta
        pin.add_observation(obs, indx=indx)
        self.assertEqual(np.max(pin.feature), 2.0)

    def test_conditions(self):
        observatory = ModelObservatory(init_load_length=1, mjd_start=survey_start_mjd())
        conditions = observatory.return_conditions()
        self.assertIsInstance(repr(conditions), str)
        self.assertIsInstance(str(conditions), str)

        step_days = 1.0

        # Number of sidereal days in a standard day
        sidereal_hours_per_day = 24 * (24.0 / 23.9344696)
        initial_lmst = float(conditions.lmst)
        conditions.mjd = conditions.mjd + step_days
        new_lmst = float(conditions.lmst)
        self.assertAlmostEqual(
            new_lmst,
            (initial_lmst + step_days * sidereal_hours_per_day) % 24,
        )

        # Check that naked conditions work
        conditions_naked = features.Conditions()
        _ = conditions_naked.__str__()

    def test_note_last_observed(self):
        note_last_observed = features.NoteLastObserved(note="test")

        observation = empty_observation()
        observation["mjd"] = 59000.0

        note_last_observed.add_observation(observation=observation)

        assert note_last_observed.feature is None

        observation["note"] = "foo"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature == observation["mjd"]

    def test_note_last_observed_with_filter(self):
        note_last_observed = features.NoteLastObserved(
            note="test",
            filtername="r",
        )

        observation = empty_observation()
        observation["mjd"] = 59000.0

        note_last_observed.add_observation(observation=observation)

        assert note_last_observed.feature is None

        observation["note"] = "foo"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"
        observation["filter"] = "g"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature is None

        observation["note"] = "test"
        observation["filter"] = "r"

        note_last_observed.add_observation(observation=observation)
        assert note_last_observed.feature == observation["mjd"]

    def test_NObservationsCurrentSeason(self):
        # Start with basic NObservationsCurrentSeason - no restrictions
        mjd_start = survey_start_mjd()
        nside = 64
        season_feature = features.NObservationsCurrentSeason(nside=nside, mjd_start=mjd_start)
        # Check that season map changes as expected with updates in Conditions
        conditions = features.Conditions(nside=nside)
        conditions.mjd = mjd_start
        season = deepcopy(season_feature.season)
        # with same time, should have same seasons.
        season_feature.season_update(conditions=conditions)
        self.assertTrue(np.all(season_feature.season == season))
        # Advancing time should update seasons in some parts of the sky
        # And leave it the same in other parts of the sky
        conditions.mjd = mjd_start + 100
        season_feature.season_update(conditions=conditions)
        self.assertTrue(np.any(season_feature.season != season))
        self.assertTrue(np.any(season_feature.season == season))
        season = deepcopy(season_feature.season)
        # Do we send a warning if time goes backwards
        # and keep the season and feature the same
        conditions.mjd = mjd_start
        with self.assertWarns(UserWarning):
            season_feature.season_update(conditions=conditions)
        self.assertTrue(np.all(season_feature.season == season))
        # And do the same for observations - check update_seasons works
        # Make some observations
        observations = [empty_observation(), empty_observation()]
        observations[0]["mjd"] = mjd_start
        observations[0]["ID"] = 0
        observations[1]["mjd"] = mjd_start + 100
        observations[1]["ID"] = 1
        season_feature = features.NObservationsCurrentSeason(nside=16, mjd_start=mjd_start)
        season = deepcopy(season_feature.season)
        season_feature.season_update(observation=observations[0])
        self.assertTrue(np.all(season_feature.season == season))
        season_feature.season_update(observation=observations[1])
        self.assertTrue(np.any(season_feature.season != season))
        self.assertTrue(np.any(season_feature.season == season))
        season = deepcopy(season_feature.season)
        # Make time go backwards .. should NOT warn but should not
        # update feature.
        # with self.assertWarns(UserWarning):
        #    season_feature.season_update(observation=observations[0])
        self.assertTrue(np.all(season_feature.season == season))

        # Now check that the _feature) works as expected when we
        # add_observation (i.e. we now care about location and healpix)
        for obs in observations:
            obs["RA"] = np.radians(30)
            obs["dec"] = np.radians(-20)
        observations[1]["mjd"] = mjd_start + 0.3
        pointing2indx = HpInLsstFov(nside=nside)
        indxs = pointing2indx(observations[0]["RA"], observations[0]["dec"])
        # first with no restrictions
        season_feature = features.NObservationsCurrentSeason(nside=nside, mjd_start=mjd_start)
        self.assertTrue(np.all(season_feature.feature == 0))
        season_feature.add_observation(observations[0], indxs)
        # Assert that we added an observation at each of the indexes
        self.assertTrue(np.all(season_feature.feature[indxs] == 1))
        # and nowhere else
        self.assertTrue(np.all(np.delete(season_feature.feature, indxs) == 0))
        # Add another observation within a day
        season_feature.add_observation(observations[1], indxs)
        # Assert that we added an observation at each of the indexes
        self.assertTrue(np.all(season_feature.feature[indxs] == 2))
        # and nowhere else
        self.assertTrue(np.all(np.delete(season_feature.feature, indxs) == 0))
        # Add an observation at a different point on the sky, but where
        # season should not turn oer for ra above yet.
        observations.append(empty_observation())
        observations[-1]["mjd"] = mjd_start + 10
        observations[-1]["RA"] = np.radians(50)
        observations[-1]["dec"] = np.radians(-20)
        observations[-1]["ID"] = len(observations) - 1
        indxs2 = pointing2indx(observations[-1]["RA"], observations[-1]["dec"])
        season_feature.add_observation(observations[-1], indxs2)
        self.assertTrue(np.all(season_feature.feature[indxs2] == 1))
        self.assertTrue(np.all(season_feature.feature[indxs] == 2))
        # Check that if we fast-forward to a future year, and add an
        # observation that everything reset to 1 or 0.
        observation = observations[0]
        observation["mjd"] = mjd_start + 365.25 * 2
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature[indxs] == 1))
        self.assertTrue(np.all(np.delete(season_feature.feature, indxs) == 0))

        # Set up to check add_observations_array - use observations from above.
        observations[0]["mjd"] = mjd_start
        observations_array = np.empty(len(observations), dtype=observations[0].dtype)
        for i, obs in enumerate(observations):
            observations_array[i] = obs
        # Build observations_hpids_array.
        # Find list of lists of healpixels
        # (should match [indxs, indxs, indxs2] from above)
        list_of_hpids = pointing2indx(observations_array["RA"], observations_array["dec"])
        indxs_list = [indxs, indxs, indxs2]
        for hi, ii in zip(list_of_hpids, indxs_list):
            self.assertTrue(set(hi) == set(ii))
        # Unravel list-of-lists (list_of_hpids) to match against observations
        hpids = []
        big_array_indx = []
        for i, indxs in enumerate(list_of_hpids):
            for indx in indxs:
                hpids.append(indx)
                big_array_indx.append(i)
        hpids = np.array(hpids, dtype=[("hpid", int)])
        # Set up format / dtype for observations_hpids_array
        names = list(observations_array.dtype.names)
        types = [observations_array[name].dtype for name in names]
        names.append(hpids.dtype.names[0])
        types.append(hpids["hpid"].dtype)
        ndt = list(zip(names, types))
        observations_hpids_array = np.empty(hpids.size, dtype=ndt)
        # Populate observations_hpid_array - big_array_indx points
        # between index in observations_array and index in hpid
        observations_hpids_array[list(observations_array.dtype.names)] = observations_array[big_array_indx]
        observations_hpids_array[hpids.dtype.names[0]] = hpids
        # Now check that adding observations to feature via
        # add_observations_array results in the same feature
        feature_obs = features.NObservationsCurrentSeason(nside=nside, mjd_start=mjd_start)
        for obs, idx in zip(observations, indxs_list):
            feature_obs.add_observation(obs, idx)
        feature_arr = features.NObservationsCurrentSeason(nside=nside, mjd_start=mjd_start)
        feature_arr.add_observations_array(observations_array, observations_hpids_array)
        self.assertTrue(np.all(feature_obs.season == feature_arr.season))
        self.assertTrue(np.all(feature_obs.feature == feature_arr.feature))

        # Check that feature works as expected when adding requirements.
        # Add seeing requirement.
        season_feature = features.NObservationsCurrentSeason(
            nside=nside, mjd_start=mjd_start, seeing_fwhm_max=1.0
        )
        observation = empty_observation()
        observation["mjd"] = mjd_start
        observation["RA"] = np.radians(30)
        observation["dec"] = np.radians(-20)
        # Don't add if seeing is too bad
        observation["FWHMeff"] = 1.5
        pointing2indx = HpInLsstFov(nside=nside)
        indxs = pointing2indx(observation["RA"], observation["dec"])
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature == 0))
        # Do add if seeing is good enough
        observation["FWHMeff"] = 0.5
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature[indxs] == 1))
        # Add filter requirement.
        season_feature = features.NObservationsCurrentSeason(
            nside=nside, mjd_start=mjd_start, seeing_fwhm_max=1.0, filtername="r"
        )
        observation["filter"] = "g"
        # Don't add if in the wrong filter
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature == 0))
        # Do add if correct filter and good seeing.
        observation["filter"] = "r"
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature[indxs] == 1))
        # Add m5 requirement.
        season_feature = features.NObservationsCurrentSeason(
            nside=nside, mjd_start=mjd_start, seeing_fwhm_max=1.0, filtername="r", m5_penalty_max=0
        )
        dark_map = dark_sky(nside)["r"]
        # Don't add if too bright
        observation["fivesigmadepth"] = dark_map[indxs].min() - 10
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature == 0))
        # Do add if faint enough
        observation["fivesigmadepth"] = dark_map[indxs].max() + 10
        season_feature.add_observation(observation, indxs)
        self.assertTrue(np.all(season_feature.feature[indxs] == 1))
        # Should test also that add_observations_array works
        # in these cases with added requirements .. but will leave it
        # to the "restore" test in test_utils.py.


if __name__ == "__main__":
    unittest.main()



import unittest

from rubin_scheduler.data import data_versions, get_data_dir


class DataTest(unittest.TestCase):
    def testData(self):
        """
        Check that basic data tools work
        """
        data_dir = get_data_dir()
        versions = data_versions()
        assert data_dir is not None
        assert versions is not None


if __name__ == "__main__":
    unittest.main()
